# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from hta.configs.config import logger
from hta.utils.utils import (
    get_kernel_type,
    IdleTimeType,
    KernelType,
    merge_kernel_intervals,
)
from plotly.subplots import make_subplots

# import statement used without the "if TYPE_CHECKING" guard will cause a circular
# dependency with trace_analysis.py causing mypy to fail and should not be removed.
if TYPE_CHECKING:
    from hta.common.trace import Trace

# This configures the threshold under which we consider gaps between
# kernels to be due to realistic delays in launching back-back kernels on the GPU


class BreakdownAnalysis:
    def __init__(self):
        pass

    @classmethod
    def get_gpu_kernel_breakdown(
        cls,
        t: "Trace",
        visualize: bool = True,
        duration_ratio: float = 0.8,
        num_kernels: int = 10,
        include_memory_kernels: bool = False,
        image_renderer="notebook",
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        GPU kernel breakdown implementation. See `get_gpu_kernel_breakdown` in `trace_analysis.py` for details.
        """
        sym_table = t.symbol_table.get_sym_table()

        all_kernel_df = pd.DataFrame(
            {
                "name": pd.Series(dtype="str"),
                "sum": pd.Series(dtype="int"),
                "max": pd.Series(dtype="int"),
                "min": pd.Series(dtype="int"),
                "std": pd.Series(dtype="float"),
                "mean": pd.Series(dtype="int"),
                "kernel_type": pd.Series(dtype="str"),
                "rank": pd.Series(dtype="int"),
            }
        )
        kernel_type_df = pd.DataFrame(
            {
                "kernel_type": pd.Series(dtype="str"),
                "sum": pd.Series(dtype="int"),
            }
        )

        kernel_type_to_analysis: List[str] = [
            KernelType.COMPUTATION.name,
            KernelType.COMMUNICATION.name,
        ]
        if include_memory_kernels:
            kernel_type_to_analysis.append(KernelType.MEMORY.name)

        kernel_per_rank: Dict[str, Dict] = defaultdict(dict)
        for rank, trace_df in t.traces.items():
            gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()
            gpu_kernels["kernel_type"] = gpu_kernels[["name"]].apply(
                lambda x: get_kernel_type(sym_table[x["name"]]), axis=1
            )
            gpu_kernels["name"] = gpu_kernels["name"].apply(lambda x: sym_table[x])

            # Create kernel type dataframe
            kernel_type_df = pd.concat(
                [
                    kernel_type_df,
                    cls._get_gpu_kernel_type_time(gpu_kernels, kernel_type_to_analysis),
                ],
                ignore_index=True,
            )

            # Create all kernel info dataframe
            for kernel_type in kernel_type_to_analysis:
                gpu_kernel_time = gpu_kernels[gpu_kernels["kernel_type"] == kernel_type]

                if kernel_type not in kernel_per_rank:
                    kernel_per_rank[kernel_type] = {}

                gpu_kernel_time = cls._aggr_gpu_kernel_time(
                    gpu_kernel_time,
                    duration_ratio=duration_ratio,
                    num_kernels=num_kernels,
                )

                kernel_per_rank[kernel_type][rank] = gpu_kernel_time

                gpu_kernel_time["kernel_type"] = kernel_type
                gpu_kernel_time["rank"] = int(rank)
                all_kernel_df = pd.concat(
                    [all_kernel_df, gpu_kernel_time], ignore_index=True
                )

        kernel_type_df = kernel_type_df.groupby(by=["kernel_type"])["sum"].agg(["sum"])
        kernel_type_df.reset_index(inplace=True)
        kernel_type_df.sort_values(
            by=["sum"], ignore_index=True, inplace=True, ascending=False
        )
        kernel_type_df["percentage"] = (
            kernel_type_df["sum"] / kernel_type_df["sum"].sum()
        ) * 100
        kernel_type_df = kernel_type_df.round({"percentage": 1})

        all_kernel_df.sort_values(
            by=["kernel_type", "name", "rank"], ignore_index=True, inplace=True
        )
        all_kernel_df.rename(
            columns={
                "sum": "sum (us)",
                "mean": "mean (us)",
                "max": "max (us)",
                "min": "min (us)",
                "std": "stddev",
            },
            inplace=True,
        )

        if visualize:  # pragma: no cover
            non_zero_kernel_df = kernel_type_df[(kernel_type_df["percentage"] > 0)]

            fig = px.pie(
                non_zero_kernel_df,
                values="percentage",
                names="kernel_type",
                height=500,
                title="Kernel Type Percentage Across All Ranks",
            )
            fig.update_layout(
                margin=dict(l=50, r=50, b=50, t=50),
                showlegend=True,
                legend=dict(yanchor="bottom", y=-0.4, xanchor="left", x=0),
            )
            fig.show(renderer=image_renderer)

            for kernel in kernel_per_rank:
                specs = []
                for count, rank in enumerate(kernel_per_rank[kernel]):
                    if count % 2 == 0:
                        specs.append([{"type": "domain"}, {"type": "domain"}])
                fig = make_subplots(
                    rows=int((len(kernel_per_rank[kernel]) + 1) / 2),
                    cols=2,
                    specs=specs,
                )
                for rank in kernel_per_rank[kernel]:
                    fig.add_trace(
                        go.Pie(
                            labels=kernel_per_rank[kernel][rank]["name"],
                            values=kernel_per_rank[kernel][rank]["sum"],
                            title=f"Rank {rank}",
                            automargin=False,
                        ),
                        int(rank / 2) + 1,
                        int(rank % 2) + 1,
                    )
                image_size_multiplier = 1 + (len(t.traces.keys())) / 2
                fig.update_layout(
                    title_text=f'Kernel type "{kernel}" - kernel distribution on each rank',
                    margin=dict(l=50, r=50, b=50, t=50),
                    showlegend=True,
                    height=400 * image_size_multiplier,
                    legend=dict(yanchor="bottom", y=-0.1, xanchor="left", x=0),
                )
                fig.show(renderer=image_renderer)

                kernel_df = all_kernel_df[all_kernel_df["kernel_type"].eq(kernel)]

                kernel_name = kernel_df["name"].unique()
                for name in kernel_name:
                    if name != "others":
                        kernel_name_df = kernel_df[kernel_df["name"].eq(name)]
                        fig = px.bar(
                            kernel_name_df,
                            x="rank",
                            y="mean (us)",
                            title=name,
                            labels={
                                "rank": "Rank",
                                "mean (us)": "Mean Duration (us)",
                            },
                            error_y=kernel_name_df["max (us)"]
                            - kernel_name_df["mean (us)"],
                            error_y_minus=kernel_name_df["mean (us)"]
                            - kernel_name_df["min (us)"],
                        )
                        fig.update_layout(
                            title_text=f'Kernel type "{kernel}" - {name}',
                            xaxis=dict(tickmode="linear", tick0=0, dtick=1),
                        )
                        fig.show(renderer=image_renderer)

        return kernel_type_df, all_kernel_df

    @classmethod
    def _get_gpu_kernel_type_time(
        cls, gpu_kernels: pd.DataFrame, kernel_type_to_analysis: List[str]
    ) -> pd.DataFrame:
        overlap_kernel_type_df = pd.DataFrame(
            {
                "status": pd.Series(dtype="str"),
                "time": pd.Series(dtype="int"),
            }
        )

        kernel_t_mapping: Dict[str, int] = defaultdict(int)
        for idx, kernel_type in enumerate(kernel_type_to_analysis):
            value = 1 << idx
            kernel_t_mapping[kernel_type] = value
            kernel_t_df = merge_kernel_intervals(
                gpu_kernels[gpu_kernels["kernel_type"].eq(kernel_type)].copy()
            )

            overlap_kernel_type_df = (
                pd.concat(
                    [
                        overlap_kernel_type_df,
                        kernel_t_df.melt(var_name="status", value_name="time").replace(
                            {"ts": value, "end": -value}
                        ),
                    ]
                )
                .sort_values(by="time")
                .reset_index(drop=True)
            )

        overlap_kernel_type_df["running"] = overlap_kernel_type_df["status"].cumsum()
        overlap_kernel_type_df["next_time"] = overlap_kernel_type_df["time"].shift(-1)
        unique_running = overlap_kernel_type_df["running"].unique()
        running_mapping: Dict[int, str] = defaultdict(str)
        for u_running in unique_running:
            if u_running > 0:
                for k_t, v_t in kernel_t_mapping.items():
                    if u_running & v_t:
                        if u_running not in running_mapping:
                            running_mapping[u_running] = k_t
                        else:
                            running_mapping[
                                u_running
                            ] = f"{running_mapping[u_running]} overlapping {k_t}"

        overlap_kernel_type_df["kernel_type"] = ""
        overlap_kernel_type_df = overlap_kernel_type_df[
            overlap_kernel_type_df["running"] > 0
        ]
        for running in running_mapping:
            overlap_kernel_type_df.loc[
                overlap_kernel_type_df["running"].eq(running), "kernel_type"
            ] = running_mapping[running]
        overlap_kernel_type_df["dur"] = (
            overlap_kernel_type_df["next_time"] - overlap_kernel_type_df["time"]
        ).astype(int)

        overlap_kernel_type_df = overlap_kernel_type_df.groupby(by=["kernel_type"])[
            "dur"
        ].agg(["sum"])
        overlap_kernel_type_df.reset_index(inplace=True)

        return overlap_kernel_type_df

    @classmethod
    def _aggr_gpu_kernel_time(
        cls,
        gpu_kernel_time: pd.DataFrame,
        duration_ratio: float = 0.8,
        num_kernels: int = 10,
    ) -> pd.DataFrame:
        gpu_kernel_time = gpu_kernel_time.groupby(by=["name"])["dur"].agg(
            ["sum", "max", "min", "mean", "std"]
        )
        gpu_kernel_time.reset_index(inplace=True)
        gpu_kernel_time = gpu_kernel_time.sort_values(
            by=["sum"], ascending=False, ignore_index=True
        )
        gpu_kernel_time["std"].fillna(0, inplace=True)

        # if there are more than num_kernels kernels, starting to aggregate kernels
        if gpu_kernel_time.shape[0] > num_kernels:
            gpu_kernel_time["cumsum"] = gpu_kernel_time["sum"].cumsum()
            quantiles = gpu_kernel_time["cumsum"].quantile(duration_ratio)
            gpu_kernel_time.loc[
                gpu_kernel_time["cumsum"] > quantiles, "name"
            ] = "others"
            gpu_kernel_time.loc[gpu_kernel_time.index >= num_kernels, "name"] = "others"
            gpu_kernel_time = gpu_kernel_time.groupby(by=["name"])["sum"].agg(
                ["sum", "max", "min", "mean", "std"]
            )
            gpu_kernel_time.reset_index(inplace=True)
            gpu_kernel_time["std"].fillna(0, inplace=True)

        return gpu_kernel_time

    @classmethod
    def _get_idle_time_for_kernels(cls, kernels_df: pd.DataFrame) -> Tuple[int, int]:
        """
        Compute idle time for given set of GPU kernels :
          returns :
            idle time (us) = kernel time - merged execution time of all kernels
            kernel time (us) = defined as the time difference between end of the
                         last kernel and start of the first kernel.
            PS: we exclude the last profiler iteration while reading trace
            so total time is exclusive of that.
        """
        merged_kernels = merge_kernel_intervals(kernels_df)
        kernel_time = merged_kernels.iloc[-1]["end"] - merged_kernels.iloc[0]["ts"]
        # differences of end - ts are commutative
        kernel_run_time = merged_kernels.end.sum() - merged_kernels.ts.sum()
        return kernel_time - kernel_run_time, kernel_time

    @classmethod
    def get_temporal_breakdown(cls, t: "Trace", visualize: bool = True) -> pd.DataFrame:
        """
        Temporal breakdown implementation. See `get_temporal_breakdown` in `trace_analysis.py` for details.
        """
        sym_table = t.symbol_table.get_sym_table()

        def idle_time_per_rank(trace_df: pd.DataFrame) -> Tuple[int, int, int, int]:
            """returns idle_time (us) , compute_time (us), non_compute_time (us), total_time (us)"""
            gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()
            idle_time, kernel_time = cls._get_idle_time_for_kernels(gpu_kernels)

            gpu_kernels["kernel_type"] = gpu_kernels[["name"]].apply(
                lambda x: get_kernel_type(sym_table[x["name"]]), axis=1
            )

            # Isolate computation kernels and merge each one of them.
            comp_kernels = merge_kernel_intervals(
                gpu_kernels[
                    gpu_kernels["kernel_type"].eq(KernelType.COMPUTATION.name)
                ].copy()
            )
            compute_time = comp_kernels.end.sum() - comp_kernels.ts.sum()
            non_compute_time = kernel_time - compute_time - idle_time

            assert idle_time <= kernel_time
            assert compute_time <= kernel_time
            assert non_compute_time >= 0

            return idle_time, compute_time, non_compute_time, kernel_time

        result: Dict[str, List[float]] = defaultdict(list)
        for rank, trace_df in t.traces.items():
            result["rank"].append(rank)
            idle_time, compute_time, non_compute_time, kernel_time = idle_time_per_rank(
                trace_df
            )
            result["idle_time(us)"].append(idle_time)
            result["compute_time(us)"].append(compute_time)
            result["non_compute_time(us)"].append(non_compute_time)
            result["kernel_time(us)"].append(kernel_time)

        result_df = pd.DataFrame(result)
        result_df["idle_time"] = (
            result_df["idle_time(us)"] / result_df["kernel_time(us)"]
        )
        result_df["idle_time_pctg"] = round(100 * result_df["idle_time"], 2)
        result_df["compute_time"] = (
            result_df["compute_time(us)"] / result_df["kernel_time(us)"]
        )
        result_df["compute_time_pctg"] = round(100 * result_df["compute_time"], 2)
        result_df["non_compute_time"] = (
            result_df["non_compute_time(us)"] / result_df["kernel_time(us)"]
        )
        result_df["non_compute_time_pctg"] = round(
            100 * result_df["non_compute_time"], 2
        )

        if visualize:  # pragma: no cover
            fig = px.bar(
                result_df,
                x="rank",
                y=["idle_time", "compute_time", "non_compute_time"],
                title="Temporal breakdown across ranks",
                labels={
                    "rank": "Rank",
                },
            )
            fig.update_layout(
                yaxis_tickformat=".2%",
                yaxis_title="Percentage",
                legend_title="Time Breakdown",
            )
            fig.show()

        return result_df[
            [
                "rank",
                "idle_time(us)",
                "compute_time(us)",
                "non_compute_time(us)",
                "kernel_time(us)",
                "idle_time_pctg",
                "compute_time_pctg",
                "non_compute_time_pctg",
            ]
        ]

    @classmethod
    def _analyze_idle_time_for_stream(
        cls,
        stream: int,
        gpu_kernels: pd.DataFrame,
        consecutive_kernel_delay: int,
        show_idle_interval_stats=False,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        """Analyze a specific CUDA stream for idle time breakdown on it.

        stream (int): CUDA stream to consider.
        gpu_kernels: a dataframe of GPU kernels in a rank.

        returns
        1) dataframe with idle time breakdown.
        1) optional dataframe showing idle interval statistics.
        """
        logger.info(f"Processing stream {stream}")
        idle_interval_stats: Optional[pd.DataFrame] = None

        gpu_kernels_s = (
            gpu_kernels[gpu_kernels.stream == stream].copy().sort_values(by="ts")
        )

        gpu_kernels_s["end_ts"] = gpu_kernels_s.ts + gpu_kernels_s.dur
        gpu_kernels_s["prev_end_ts"] = gpu_kernels_s.end_ts.shift(1)
        gpu_kernels_s["idle_interval"] = (
            gpu_kernels_s["ts"] - gpu_kernels_s["prev_end_ts"]
        )

        # Default idle time category
        gpu_kernels_s["idle_category"] = IdleTimeType.OTHER.value

        """
        Host wait:
            If the current kernel's runtime started after previous kernel's end time
            this means the Host/CPU was not enqueuing kernels fast enough.
        CPU  Runtime 0             Runtime 1
        GPU     |--------Kernel 0      |-----------Kernel 1
        """
        is_host_wait = gpu_kernels_s["ts_runtime"] > gpu_kernels_s["prev_end_ts"]
        gpu_kernels_s.loc[is_host_wait, "idle_category"] = IdleTimeType.HOST_WAIT.value

        """
        Kernel wait:
            If the gap between kernels is below a threshold the idle time is
            likely due to the overhead for launching kernels.
        """
        is_kernel_kernel_delay = ~is_host_wait & (
            gpu_kernels_s["idle_interval"] < consecutive_kernel_delay
        )
        gpu_kernels_s.loc[
            is_kernel_kernel_delay, "idle_category"
        ] = IdleTimeType.KERNEL_WAIT.value

        gpu_kernels_groupby = gpu_kernels_s.groupby("idle_category")
        if show_idle_interval_stats:
            logger.info(
                f"Computing descriptive statistics for idle time intervals on stream {stream}:"
            )
            idle_interval_stats = gpu_kernels_groupby.idle_interval.describe()
            idle_interval_stats.insert(0, "stream", stream)

        result = pd.DataFrame(gpu_kernels_groupby.idle_interval.sum())
        total_idle_time = result.idle_interval.sum()
        result["stream"] = stream
        result["idle_time_ratio"] = result["idle_interval"] / total_idle_time
        result.rename(columns={"idle_interval": "idle_time"}, inplace=True)
        return result, idle_interval_stats

    @classmethod
    def get_idle_time_breakdown(
        cls,
        t: "Trace",
        consecutive_kernel_delay: int,
        rank: int = 0,
        streams: Optional[List[int]] = None,
        visualize: bool = True,
        visualize_pctg: bool = True,
        show_idle_interval_stats=False,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        """
        Breakdown Idle time by host wait, kernel wait and other categories. See full description in trace_analysis.py

        consecutive_kernel_delay (int): configures the threshold under which we consider gaps between
           kernels to be due to realistic delays in launching back-back kernels on the GPU. Time is in ns.
        rank (int): the rank to analyze
        streams (List[int]): list of streams to provide analysis for.
            Defaults to all streams.
        visualize (bool): show the visualization chart or not (default = True).
        visualize_pctg (bool): show relative percentage across streams (default = True).
        show_idle_interval_stats (bool): prints statistics of the idle intervals like the min, max
           and median of idle intervals between kernels on a CUDA stream, also broken down by
           the idleness category (default = False).
        """
        trace_df: pd.DataFrame = t.get_trace(rank)
        gpu_kernels_pre = (
            trace_df[trace_df["stream"].ne(-1)].copy().set_index("index_correlation")
        )

        # correlate with the runtime event whenever possible
        gpu_kernels = gpu_kernels_pre.join(
            trace_df[["ts", "index"]], on="index_correlation", rsuffix="_runtime"
        )

        if streams is None or len(streams) == 0:
            streams = list(gpu_kernels.stream.unique())

        result_list: List[pd.DataFrame] = []
        interval_stats_list: List[pd.DataFrame] = []

        for stream in streams:
            breakdown_df, idle_interval_df = cls._analyze_idle_time_for_stream(
                stream,
                gpu_kernels,
                consecutive_kernel_delay,
                show_idle_interval_stats,
            )
            result_list.append(breakdown_df)
            if idle_interval_df is not None:
                interval_stats_list.append(idle_interval_df)

        result_df = pd.concat(result_list)

        idle_category_name_map = {
            member.value: name.lower()
            for name, member in IdleTimeType.__members__.items()
        }
        result_df.rename(mapper=idle_category_name_map, axis=0, inplace=True)
        result_df.reset_index(inplace=True)

        if visualize:  # pragma: no cover
            result_df["stream"] = result_df.stream.astype(str)
            ycol = "idle_time_ratio" if visualize_pctg else "idle_time"
            fig = px.bar(
                result_df,
                x="stream",
                y=ycol,
                color="idle_category",
                hover_data=["idle_time", "idle_time_ratio"],
                title=f"Idle time breakdown on rank {rank} per CUDA stream",
            )
            if visualize_pctg:
                fig.update_layout(
                    yaxis_tickformat=".2%",
                    yaxis_title="Percentage",
                    legend_title="Idle Time Breakdown",
                )
            else:
                fig.update_layout(
                    yaxis_title="Idle time (us)", legend_title="Idle Time Breakdown"
                )
            fig.show()

        result_df["rank"] = rank
        interval_stats_df = (
            pd.concat(interval_stats_list).round(2)
            if show_idle_interval_stats
            else None
        )
        if interval_stats_df is not None:
            # add rank column to the starting
            interval_stats_df.insert(0, "rank", rank)
            interval_stats_df.rename(
                mapper=idle_category_name_map, axis=0, inplace=True
            )

        result_df = result_df[
            ["rank", "stream", "idle_category", "idle_time", "idle_time_ratio"]
        ].round(2)

        return result_df, interval_stats_df



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import setuptools
from docs.conf import find_version


def fetch_requirements():
    with open("requirements.txt") as f:
        reqs = f.read().strip().split("\n")
    return reqs


setuptools.setup(
    name="HolisticTraceAnalysis",
    description="A python library for analyzing PyTorch Profiler traces",
    version=find_version("hta/version.py"),
    url="https://github.com/facebookresearch/HolisticTraceAnalysis",
    python_requires=">=3.8",
    author="Meta Platforms Inc.",
    author_email="todo@meta.com",
    license="MIT",
    install_requires=fetch_requirements(),
    include_package_data=True,
    packages=setuptools.find_packages(
        include=["hta*"]
    ),  # Only include code within hta.
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: OS Independent",
    ],
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Any, Dict

import pandas as pd

from hta.trace_diff import DeviceType, LabeledTrace, TraceDiff


class TraceCompareTestCase(unittest.TestCase):
    def setUp(self):
        self.base_trace: Dict[str, Any] = {
            "label": "Base Trace",
            "trace_dir": "tests/data/trace_compare/base",
            "ranks": [1],
            "iterations": [1009, 1010],
            "test_rank": 1,
            "test_iteration": 1010,
            "test_num_cpu_events": 6,
            "test_num_gpu_events": 1,
            "test_user_annotation_summary": {
                "counts": {
                    "ProfilerStep#1010": 1,
                    "aten::as_strided": 1,
                    "aten::empty": 1,
                    "cudaEventQuery": 1,
                    "cudaLaunchKernel": 1,
                    "nccl:all_reduce": 1,
                },
                "total_duration": {
                    "ProfilerStep#1010": 122149,
                    "aten::as_strided": 0,
                    "aten::empty": 0,
                    "cudaEventQuery": 4,
                    "cudaLaunchKernel": 4,
                    "nccl:all_reduce": 67,
                },
            },
        }
        self.test_trace: Dict[str, Any] = {
            "label": "Test Trace",
            "trace_dir": "tests/data/trace_compare/test",
            "ranks": [1],
            "iterations": [1009, 1010],
            "test_rank": 1,
            "test_iteration": 1010,
            "counts_change_categories": {
                "cpu": {"=": 5, "+": 3, "-": 1},
                "gpu": {"-": 1},
            },
            "ops_diff": {
                "added": ["AddmmBackward0", "All2All_Pooled_Req", "CatBackward0"],
                "deleted": ["nccl:all_reduce"],
                "increased": [],
                "decreased": [],
                "unchanged": [
                    "ProfilerStep#1010",
                    "aten::as_strided",
                    "aten::empty",
                    "cudaEventQuery",
                    "cudaLaunchKernel",
                ],
            },
        }

    def test_labeled_trace(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        self.assertListEqual(base_t.ranks(), self.base_trace["ranks"])
        self.assertListEqual(base_t.iterations(), self.base_trace["iterations"])

    def test_labeled_trace_random_label(self) -> None:
        base_t = LabeledTrace(trace_dir=self.base_trace["trace_dir"])
        self.assertTrue(base_t.label.startswith("t"))

    def test_extract_ops(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        cpu_events: pd.DataFrame = base_t.extract_ops(
            self.base_trace["test_rank"],
            self.base_trace["test_iteration"],
            DeviceType.CPU,
        )
        self.assertEqual(cpu_events.shape[0], self.base_trace["test_num_cpu_events"])
        gpu_events: pd.DataFrame = base_t.extract_ops(
            self.base_trace["test_rank"],
            self.base_trace["test_iteration"],
            DeviceType.GPU,
        )
        self.assertEqual(gpu_events.shape[0], self.base_trace["test_num_gpu_events"])

    def test_get_ops_summary(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        df: pd.DataFrame = base_t.get_ops_summary(
            base_t.extract_ops(
                self.base_trace["test_rank"],
                self.base_trace["test_iteration"],
                DeviceType.CPU,
            )
        )
        self.assertDictEqual(
            df.sort_values(by="name")[["name", "counts", "total_duration"]]
            .set_index("name")
            .to_dict(),
            self.base_trace["test_user_annotation_summary"],
        )

    def test_compare_trace(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        base_rank = self.base_trace["test_rank"]
        base_iteration = self.base_trace["test_iteration"]

        test_t = LabeledTrace(
            label=self.test_trace["label"], trace_dir=self.test_trace["trace_dir"]
        )
        test_rank = self.test_trace["test_rank"]
        test_iteration = self.test_trace["test_iteration"]

        cpu_df = TraceDiff.compare_traces(
            base_t,
            test_t,
            base_rank,
            test_rank,
            base_iteration,
            test_iteration,
            device_type=DeviceType.CPU,
            use_short_name=True,
        )
        self.assertDictEqual(
            cpu_df["counts_change_categories"].value_counts().to_dict(),
            self.test_trace["counts_change_categories"]["cpu"],
        )

        gpu_df = TraceDiff.compare_traces(
            base_t,
            test_t,
            base_rank,
            test_rank,
            base_iteration,
            test_iteration,
            device_type=DeviceType.GPU,
            use_short_name=True,
        )
        self.assertDictEqual(
            gpu_df["counts_change_categories"].value_counts().to_dict(),
            self.test_trace["counts_change_categories"]["gpu"],
        )

    def test_ops_diff(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        base_rank = self.base_trace["test_rank"]
        base_iteration = self.base_trace["test_iteration"]

        test_t = LabeledTrace(
            label=self.test_trace["label"], trace_dir=self.test_trace["trace_dir"]
        )
        test_rank = self.test_trace["test_rank"]
        test_iteration = self.test_trace["test_iteration"]

        diff = TraceDiff.ops_diff(
            base_t,
            test_t,
            base_rank,
            test_rank,
            base_iteration,
            test_iteration,
            device_type=DeviceType.CPU,
        )
        self.assertDictEqual(diff, self.test_trace["ops_diff"])

    def test_visualize(self) -> None:
        base_t = LabeledTrace(
            label=self.base_trace["label"], trace_dir=self.base_trace["trace_dir"]
        )
        base_rank = self.base_trace["test_rank"]
        base_iteration = self.base_trace["test_iteration"]

        test_t = LabeledTrace(
            label=self.test_trace["label"], trace_dir=self.test_trace["trace_dir"]
        )
        test_rank = self.test_trace["test_rank"]
        test_iteration = self.test_trace["test_iteration"]

        cpu_df = TraceDiff.compare_traces(
            base_t,
            test_t,
            base_rank,
            test_rank,
            base_iteration,
            test_iteration,
            device_type=DeviceType.CPU,
            use_short_name=True,
        )
        self.assertIsNone(TraceDiff.visualize_counts_diff(cpu_df, show_image=False))
        self.assertIsNone(TraceDiff.visualize_duration_diff(cpu_df, show_image=False))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import logging.config
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union

import hta
from hta.configs.default_values import DEFAULT_CONFIG_FILENAME

ConfigValue = Union[None, bool, int, float, str, Dict[str, Any], List[Any], Set[Any]]


def setup_logger(config_file: str = "logging.config") -> logging.Logger:
    global logger
    if config_file:
        log_filepath = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "logging.config"
        )
        logging.config.fileConfig(log_filepath)
        logger = logging.getLogger("hta")
    elif logger is None:
        logger = logging.getLogger()
    return logger


logger: logging.Logger = setup_logger()


class HtaConfig:
    """
    A container for customizing the trace analyzer configurations.
    """

    @staticmethod
    def get_default_paths() -> List[str]:
        """
        Get the list of default config file paths.

        Returns:
            A list of default config file paths.

        Notes:
        To support the customization of trace analysis, hta reads configurations from
            a list of config file paths in the following order:
            1. hta/configs/trace_analyzer.json
            2. ~/.hta/trace_analyzer.json
            3. <current directory>/trace_analyzer.json
            4. a user specified, instance-specific configuration file path.

        Loading configurations from the subsequent configuration file will update previous configurations.

        If a configuration file in the above list doesn't exist, hta will skip such configuration file.
        As a class method, this function does not return the user defined config file path.
        """
        return [
            str(
                Path(hta.__file__)
                .parent.joinpath("configs")
                .joinpath(DEFAULT_CONFIG_FILENAME)
            ),
            str(Path.home().joinpath(".hta").joinpath(DEFAULT_CONFIG_FILENAME)),
            str(Path.cwd().joinpath(DEFAULT_CONFIG_FILENAME)),
        ]

    def __init__(
        self,
        config_file_path: Optional[str] = None,
        load_default_paths: Optional[bool] = True,
    ):
        """
        Constructor of the HtaConfig class.

        Args:
             config_file_path (str): a user provided config file path.
             load_default_paths (bool) : control whether the analyzer should use available default configuration files.
        """
        self.config_file_paths: List[str] = (
            HtaConfig.get_default_paths() if load_default_paths else []
        )
        if config_file_path:
            self.config_file_paths.append(config_file_path)

        self.config: Dict[str, Any] = {}
        for path in self.config_file_paths:
            if Path(path).exists() and os.access(path, os.R_OK):
                with open(path) as fp:
                    cfg = json.load(fp)
                    self.config.update(cfg)

        logging_config_file = self.get_config("hta.defaults.logging.config_file")
        if isinstance(logging_config_file, str):
            setup_logger(logging_config_file)
        else:
            setup_logger()

    def get_config_file_paths(self) -> List[str]:
        """
        Return the available config file paths
        """
        return [
            path
            for path in self.config_file_paths
            if Path(path).exists() and os.access(path, os.R_OK)
        ]

    def get_config(
        self, dot_path: Optional[str] = None, default_value: ConfigValue = None
    ) -> ConfigValue:
        """
         Return the value for a configuration key <doct_keys>.

        Args:
             dot_path (str) : a dot representation of a configuration path. For example, `a.b.c` will search the
             configuration for self.configs['a']['b']['c']. If dot_path is None, then the method will return
                 the entire set of configurations.
             default_value (ConfigValue) : return value if the dot_path doesn't exist in the configuration.

         Returns:
             The configuration value corresponding to the dot_path.

         Notes:
             HtaConfig represents its configuration using a dictionary which can have multiple layers.
             For easy use, this method allows the user to specify the search path with a dot string and then splits
             the dot string into a list of keys to navigate the configuration search.
        """
        cfg = self.config
        if dot_path is None:
            return self.config

        paths = dot_path.split(".")
        found = False
        for i, path in enumerate(paths):
            if cfg is None:
                break
            if isinstance(cfg, Dict) and path in cfg:
                cfg = cfg[path]
                if i == len(paths) - 1:
                    found = True
            else:
                break

        if found:
            return cfg
        else:
            return default_value

    def show(self):
        print(json.dumps(self.config, indent=4, sort_keys=True))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import multiprocessing as mp
from enum import Enum
from pathlib import Path
from typing import List

import pandas as pd
import psutil


class KernelType(Enum):
    COMMUNICATION = 0
    MEMORY = 1
    COMPUTATION = 2


class IdleTimeType(Enum):
    HOST_WAIT = 0
    KERNEL_WAIT = 1
    OTHER = 2


def normalize_path(path: str) -> str:
    """
    Convert a Linux path to Python path.

    Args:
        path (str) : a path acceptable by the OS.

    Returns:
        A path supported by Python.
    """
    if path.startswith("./"):
        path2 = path[2:]
        if len(path2) > 0:
            normalized_path = str(Path.cwd().joinpath(path2))
        else:
            normalized_path = str(Path.cwd())
    elif path.startswith("~/"):
        path2 = path[2:]
        if len(path2) > 0:
            normalized_path = str(Path.home().joinpath(path2))
        else:
            normalized_path = str(Path.home())
    else:
        normalized_path = path
    return normalized_path


def is_comm_kernel(name: str) -> bool:
    """
    Check if a given GPU kernel is a communication kernel.

    Args:
        name (str): name of the GPU kernel.

    Returns:
        A boolean indicating if the kernel is a communication kernel.
    """
    return "ncclKernel" in name


def is_memory_kernel(name: str) -> bool:
    """
    Check if a given GPU kernel is a memory kernel.

    Args:
        name (str): name of the GPU kernel.

    Returns:
        A boolean indicating if the kernel is an IO kernel.
    """
    return "Memcpy" in name or "Memset" in name


def get_kernel_type(name: str) -> str:
    if is_comm_kernel(name):
        return KernelType.COMMUNICATION.name
    elif is_memory_kernel(name):
        return KernelType.MEMORY.name
    else:
        return KernelType.COMPUTATION.name


def get_memory_kernel_type(name: str) -> str:
    """Memcpy Type is basically a prefix of the kernel name ~ Memcpy DtoH"""
    if name[:6] == "Memset":
        return "Memset"
    if name[:6] != "Memcpy":
        return "Memcpy Unknown"
    prefix_size = 11  # len("Memcpy DtoH")
    return name[:prefix_size]


def merge_kernel_intervals(kernel_df: pd.DataFrame) -> pd.DataFrame:
    """
    Merge all kernel intervals in the given dataframe such that there are no overlapping.
    """
    kernel_df.sort_values(by="ts", inplace=True)
    kernel_df["end"] = kernel_df["ts"] + kernel_df["dur"]
    # Operators within the same group need to be merged together to form a larger interval.
    kernel_df["group"] = (kernel_df["ts"] > kernel_df["end"].shift().cummax()).cumsum()
    kernel_df = (
        kernel_df.groupby("group", as_index=False)
        .agg({"ts": min, "end": max})
        .drop(["group"], axis=1)
        .sort_values(by="ts")
    )
    return kernel_df


def shorten_name(name: str) -> str:
    """Shorten a long operator/kernel name.

    The CPU operator and CUDA kernel name in the trace can be too long to follow.
    This utility removes the functional arguments, template arguments, and return values
    to make the name easy to understand.
    """
    s: str = name.replace("->", "")
    stack: List[str] = []
    for c in s:
        if c == ">":  # match generic template arguments
            while len(stack) and stack[-1] != "<":
                stack.pop()

            if len(stack) > 0 and stack[-1] == "<":
                stack.pop()
        elif c == ")":  # match arguments or comments
            while len(stack) and stack[-1] != "(":
                stack.pop()
            if len(stack) > 0 and stack[-1] == "(":
                stack.pop()
        else:
            stack.append(c)
    return "".join(stack).split(" ")[-1]


def flatten_column_names(df: pd.DataFrame) -> None:
    """Flatten a DataFrame's a multiple index column names to a single string"""
    if isinstance(df.columns, pd.MultiIndex):
        df.columns = ["_".join(col).rstrip("_") for col in df.columns]


def get_mp_pool_size(obj_size: int, num_objs: int) -> int:
    """
    Estimate the maximum pool size for multiprocessing

    Args:
        obj_size (int): the size of objects to be processed
        num_objs (int): the total number of objects to be processed

    Returns:
        int
            the recommend pool size
    """
    free_mem = psutil.virtual_memory().available
    # Leave 20% buffer for system and other processes
    max_np = int(0.8 * free_mem / obj_size)
    return min(max_np, num_objs, mp.cpu_count())



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from collections import namedtuple
from typing import List

from hta.utils.utils import shorten_name


class TestTracUtils(unittest.TestCase):
    def test_shorten_name(self) -> None:
        TC = namedtuple("TC", ["input", "expect_result"])
        test_cases: List[TC] = [
            TC("a", "a"),
            TC("aten::empty", "aten::empty"),
            TC("a(1)", "a"),
            TC("a<int>(b)", "a"),
            TC(
                "void at::native::vectorized_elementwise_kernel<4, at::native::CUDAFunctor_add<int>, at::detail::Array<char*, 3> >(int, at::native::CUDAFunctor_add<int>, at::detail::Array<char*, 3>)",
                "at::native::vectorized_elementwise_kernel",
            ),
        ]

        for tc in test_cases:
            self.assertEqual(shorten_name(tc.input), tc.expect_result)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import gzip
import json
import multiprocessing as mp
import os
import queue
import re
import sys
import time
import tracemalloc
from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd

from hta.common.trace_file import get_trace_files
from hta.configs.config import logger
from hta.configs.default_values import DEFAULT_TRACE_DIR
from hta.utils.utils import get_mp_pool_size, normalize_path

MetaData = Dict[str, Any]
PHASE_COUNTER: str = "C"


class _SymbolCollector:
    """
    To support a multiprocessing version of symbol table update, we use _SymbolCollector to a shared queue
    to collect all the symbols and then use a single process to merge the results. A _SymbolCollector object
    implements a function object so that it allow the multiple processes to share the data.
    """

    def __init__(self, q: queue.Queue[Any]):
        self.queue = q

    def __call__(self, symbols: Iterable[str]) -> None:
        for s in symbols:
            self.queue.put(s)


# We use a TraceSymbolTable to store the bidirectional symbol<-->id mapping for each Trace object.
# This table will be shared among all the ranks to encode/decode the symbols in their data frames.
class TraceSymbolTable:
    """
    TraceSymbolTable stores the bidirectional symbol<-->id mapping for all traces.
    Because of potential races caused by multiprocessing, we serialize updates to the
    TraceSymbolTable using a synchronize.lock.

    We assume all read operations to this table by a Trace object occur after it adds all its symbols.
    Therefore, there is no need to lock the read access.

    Attributes:
        sym_table (List[str]) : a list of symbols.
        sym_index (Dict[str, int]) : a map from symbol to ID.
    """

    def __init__(self):
        self.sym_table: List[str] = []
        self.sym_index: Dict[str, int] = {}

    def add_symbols(self, symbols: Iterable[str]) -> None:
        for s in symbols:
            if s not in self.sym_index:
                idx = len(self.sym_table)
                self.sym_table.append(s)
                self.sym_index[s] = idx

    def add_symbols_mp(self, symbols_list: List[Iterable[str]]) -> None:
        m = mp.Manager()
        shared_queue: queue.Queue[Any] = m.Queue()
        collector = _SymbolCollector(shared_queue)

        with mp.get_context("spawn").Pool(
            min(mp.cpu_count(), len(symbols_list))
        ) as pool:
            pool.map(collector, symbols_list)
            pool.close()
            pool.join()

        all_symbols = []
        while not shared_queue.empty():
            all_symbols.append(shared_queue.get())
        self.add_symbols(all_symbols)

    def get_sym_id_map(self) -> Dict[str, int]:
        return self.sym_index

    def get_sym_table(self) -> List[str]:
        return self.sym_table


def parse_trace_dict(trace_file_path: str) -> Dict[str, Any]:
    """
    Parse a raw trace file into a dictionary.

    Args:
        trace_file_path (str) : the path to a trace file.

    Returns:
        A dictionary representation of the trace.
    """
    t_start = time.perf_counter()
    trace_record: Dict[str, Any] = {}
    if trace_file_path.endswith(".gz"):
        with gzip.open(trace_file_path, "rb") as fh:
            trace_record = json.loads(fh.read())
    elif trace_file_path.endswith(".json"):
        with open(trace_file_path, "r") as fh2:
            trace_record = json.loads(fh2.read())
    else:
        raise ValueError(
            f"expect the value of trace_file ({trace_file_path}) ends with '.gz' or 'json'"
        )
    t_end = time.perf_counter()
    logger.info(f"Parsed {trace_file_path} time = {(t_end - t_start):.2f} seconds ")
    return trace_record


def compress_df(df: pd.DataFrame) -> Tuple[pd.DataFrame, TraceSymbolTable]:
    """
    Compress a Dataframe to reduce its memory footprint.

    Args:
        df (pd.DataFrame): the input DataFrame

    Returns:
        Tuple[pd.DataFrame, TraceSymbolTable]
            The first item is the compressed dataframe.
            The second item is the local symbol table specific to this dataframe.
    """
    # assign an index to each event
    df.reset_index(inplace=True)
    df["index"] = pd.to_numeric(df["index"], downcast="integer")

    # drop rows with null values
    df.dropna(axis=0, subset=["dur", "cat"], inplace=True)
    df.drop(df[df["cat"] == "Trace"].index, inplace=True)

    # drop columns
    columns_to_drop = {"ph", "id", "bp", "s"}.intersection(set(df.columns))
    df.drop(list(columns_to_drop), axis=1, inplace=True)

    # extract arguments; the argument list needs to update when more arguments are used in the analysis.
    args_to_keep = {
        "stream",
        "correlation",
        "Trace iteration",
        "memory bandwidth (GB/s)",
    }
    # performance counters appear as args
    if "cuda_profiler_range" in df.cat.unique():
        counter_names = set.union(
            *[set(d.keys()) for d in df[df.cat == "cuda_profiler_range"]["args"].values]
        )
        args_to_keep = args_to_keep.union(counter_names)

    for arg in args_to_keep:
        df[arg] = df["args"].apply(
            lambda row: row.get(arg, -1) if isinstance(row, dict) else -1
        )
    df.drop(["args"], axis=1, inplace=True)
    df.rename(columns={"memory bandwidth (GB/s)": "memory_bw_gbps"}, inplace=True)

    # create a local symbol table
    local_symbol_table = TraceSymbolTable()
    symbols = set(df["cat"].unique()).union(set(df["name"].unique()))
    local_symbol_table.add_symbols(symbols)

    sym_index = local_symbol_table.get_sym_id_map()
    for col in ["cat", "name"]:
        df[col] = df[col].apply(lambda s: sym_index[s])

    # data type downcast
    for col in df.columns:
        df[col] = pd.to_numeric(df[col], errors="coerce", downcast="integer")

    return df, local_symbol_table


def transform_correlation_to_index(df: pd.DataFrame) -> pd.DataFrame:
    """Transform correlation to index_correlation and add a index_correlation column to df.

    The correlation in the trace is a reference ID which links a Cuda kernel launch
    and the cuda kernel. Because the correlation is not an index, using `correction` to find
    the corresponding Cuda launch event for a Cuda kernel requires searching the dataframe.

    The index_correlation column maps the correction to the actual index of the linked events
    so that finding the lined events will be easy. The values of `index_correlation` fall into
    three cases:
    index_correlation = -1 : the event does not link to another event
    index_correlation = 0  : the event is cuda related but the linked event is not shown in trace.
    index_correlation > 0  : the event links to another event whose index is index_correlation.

    The transform can be illustrated as follows:

    Given the following input DataFrame <df>:

    | index | stream | cat | s_cat  | correlation |
    ===============================================
    | 675   | 7      | 248 | Kernel | 278204204   |
    | 677   | -1     |   9 | Runtime| 278204204   |

    After calling `transform_correlation_to_index(df)`, <df> will become:

    | index | stream | cat | s_cat  | correlation |  index_correlation |
    ====================================================================
    | 675   | 7      | 248 | Kernel | 278204204   |       677          |
    | 677   | -1     |   9 | Runtime| 278204204   |       675          |

    This function changes the input DataFrame by adding a new column `index_correlation`.

    Example use: find the kernels of all Cuda Launch
    cuda_launch_events = df.loc[df['cat'].eq(9) & df['index_correlation'].gt(0)]
    cuda_kernel_indices = cuda_launch_events["index_correlation"]
    df.loc[cuda_kernel_indices]

    Args:
        df (pd.DataFrame): the input DataFrame

    Returns:
        pd.DataFrame: the transformed DataFrame with a index_correlation column.

    Affects:
        This function adds a index_correlation column to df.
    """
    if "correlation" not in df.columns:
        return df
    corr_df = df.loc[df["correlation"].ne(-1), ["index", "correlation", "stream"]]
    on_cpu = corr_df.loc[df["stream"].eq(-1)]
    on_gpu = corr_df.loc[df["stream"].ne(-1)]
    merged_cpu_idx = on_cpu.merge(on_gpu, on="correlation", how="inner")
    merged_gpu_idx = on_gpu.merge(on_cpu, on="correlation", how="inner")
    matched = pd.concat([merged_cpu_idx, merged_gpu_idx], axis=0)[
        ["index_x", "index_y"]
    ].set_index("index_x")

    corr_index_map: Dict[int, int] = matched["index_y"].to_dict()

    def _set_corr_index(row):
        idx = row.name
        if idx in corr_index_map:
            return corr_index_map[idx]
        elif df.loc[idx, "correlation"] == -1:
            return -1
        else:
            return 0

    df["index_correlation"] = df.apply(_set_corr_index, axis=1)
    df["index_correlation"] = pd.to_numeric(df["index_correlation"], downcast="integer")

    return df


def add_iteration(df: pd.DataFrame, symbol_table: TraceSymbolTable) -> pd.DataFrame:
    """
    Add an iteration column to the DataFrame <df>.

    This function extracts the trace iteration number from the `ProfilerStep` annotation
    in the name column and then apply the following logic to determine which iteration
    an event belongs:
    1. For events on the CPU devices, an event `e` will have the same iteration number
       as the event ProfilerStep#<iter> `s` if  `s.ts <= e.ts < s.ts + s.dur`.
    2. For events on the GPU devices, an event `e` will have the same iteration number
       as its correlated event (linked by index_correlation or correlation).
    3. For other events whose iteration number can not be determined with the above two
       steps, set the iteration number to -1.

    Args:
        df: a DataFrame representation of a trace.
        sym_tab: the TraceSymbolTable for the trace

    Returns:
        A DataFrame with the profiler steps information.

    Note:
        This function will change the input DataFrame by adding the iteration number column.
    """
    s_map = pd.Series(symbol_table.sym_index)
    s_tab = pd.Series(symbol_table.sym_table)
    profiler_step_ids = s_map[s_map.index.str.startswith("ProfilerStep")]
    profiler_step_ids.sort_index()

    def _extract_iter(profiler_step_name: int) -> int:
        """Convert a profiler_step_name to an iteration number. For example, 168 (string name: ProfilerStep#15) ==> 15"""
        s = s_tab[profiler_step_name]
        m = re.match(r"ProfilerStep\s*#\s*(\d+)", s)
        return int(m.group(1))

    # Extract the profiler steps
    profiler_steps = df.loc[df["name"].isin(profiler_step_ids.values)]
    profiler_steps = profiler_steps[["ts", "dur", "name"]].copy()
    profiler_steps["s_name"] = profiler_steps["name"].apply(_extract_iter)
    profiler_steps["iter"] = profiler_steps["name"].apply(lambda idx: s_tab[idx])

    profiler_steps_array = profiler_steps.to_numpy()

    def _get_profiler_step(ts: int) -> int:
        """ "determine which profiler step a given timestamp <ts> falls into"""
        iter = -1
        for step in profiler_steps_array:
            if step[0] <= ts < step[0] + step[1]:
                iter = step[3]
        return iter

    # Update the trace iteration column
    df.loc[df["stream"].lt(0), "iteration"] = df["ts"].apply(_get_profiler_step)
    df.loc[df["stream"].gt(0), "iteration"] = df["index_correlation"].apply(
        lambda x: df.loc[x, "iteration"] if x > 0 else -1
    )
    df["iteration"] = pd.to_numeric(df["iteration"], downcast="integer")

    return profiler_steps


def parse_trace_dataframe(
    trace_file_path: str,
) -> Tuple[MetaData, pd.DataFrame, TraceSymbolTable]:
    """parse a single trace file into a meat test_data dictionary and a dataframe of events.
    Args:
        trace_file_path (str): The path to a trace file. When the trace_file is a relative path.
            This method combines the object's trace_path with trace_file to get the full path of the trace file.
    Returns:
        Tuple[MetaData, pd.DataFrame, TraceSymbolTable]
            The first item is the trace's metadata;
            The second item is the dataframe representation of the trace's events.
            The third item is the symbol table to encode the symbols of the trace.


    Raises:
        OSError when the trace file doesn't exist or current process has no permission to access it.
        JSONDecodeError when the trace file is not a valid JSON document.
        ValueError when the trace_file doesn't end with ".gz" or "json".
    """
    t_start = time.perf_counter()
    trace_record = parse_trace_dict(trace_file_path)

    meta: Dict[str, Any] = {k: v for k, v in trace_record.items() if k != "traceEvents"}
    df: pd.DataFrame = pd.DataFrame()
    if "traceEvents" in trace_record:
        df = pd.DataFrame(trace_record["traceEvents"])
        df, local_symbol_table = compress_df(df)
        transform_correlation_to_index(df)
        add_iteration(df, local_symbol_table)

    t_end = time.perf_counter()
    logger.debug(
        f"Parsed {trace_file_path} in {(t_end - t_start):.2f} seconds; current PID:{os. getpid()}"
    )
    return meta, df, local_symbol_table


class Trace:
    """
    A container for the traces collected for a distributed ML training job.

    An ML training job can have multiple trace collections. Each of those trace collections maps to
    one Trace object.


    Attributes:
        trace_path (str) : the path to the folder where the collected raw traces are stored. In other words,
            `trace_path = normalize_path(base_trace_dir)`.
        trace_files (Dict[int, str]) : a dictionary that maps the rank of a job's trainer to its trace file.
        traces (Dict[int, pd.DataFrame) : a dictionary that maps the rank of a job's trainer to its trace data.
        meta_data (Dict[int, MetaData) : a dictionary that maps the rank of a job's trainer to its meta_ata.
        symbol_table (TraceSymbolTable) : a symbol table used to encode the symbols in the trace.
        is_parsed (bool) : a flag indicting whether the trace is parsed or not.
    """

    def __init__(
        self,
        trace_files: Optional[Dict[int, str]] = None,
        trace_dir: str = DEFAULT_TRACE_DIR,
    ) -> None:
        """
        The constructor of a Trace object.
        Args:
            trace_files: Dict[int, str] : a map from rank to trace file names. The trace file names can be either
                relative to the path `trace_path` or absolute file paths.
            trace_dir (str) : a path used to derive `trace_path = normalize_path(trace_dir)`.

        Raises:
            AssertionError
        """
        self.trace_path: str = normalize_path(trace_dir)
        logger.info(f"{self.trace_path}")
        self.trace_files: Dict[int, str] = (
            trace_files if trace_files is not None else get_trace_files(self.trace_path)
        )
        logger.debug(self.trace_files)
        self.traces: Dict[int, pd.DataFrame] = {}
        self.symbol_table = TraceSymbolTable()
        self.meta_data: Dict[int, MetaData] = {}
        self.min_ts: int = 0

        self._normalize_trace_filenames()
        assert self._validate_trace_files()
        logger.debug(f"trace_path={self.trace_path}")
        logger.debug(f"# trace_files={len(self.trace_files)}")
        if len(self.trace_files) > 0:
            rank = next(iter(self.trace_files))
            trace_file = self.trace_files[rank]
            logger.debug(f"trace_files[{rank}] = {trace_file}")
        self.is_parsed = False

    def load_traces(self) -> None:
        if self.is_parsed:
            logger.warning("Traces are already parsed and loaded!")
            return
        self.parse_traces()
        self.align_and_filter_trace()
        for rank, trace_df in self.traces.items():
            df = self.traces[rank].set_index("index", drop=False)
            df.index.names = [None]
            self.traces[rank] = df
        self.is_parsed = True

    def parse_single_rank(self, rank: int) -> None:
        """
        Parse the trace for a given rank.

        Args:
            rank (int) : an integer representing the rank of a trainer
        """
        if rank in self.trace_files:
            trace_filepath = self.trace_files[rank]
            (
                self.meta_data[rank],
                self.traces[rank],
                local_symbol_table,
            ) = parse_trace_dataframe(trace_filepath)
            # update the global symbol table
            self.symbol_table.add_symbols(local_symbol_table.get_sym_table())
            # fix the encoding of the data frame
            local_table = local_symbol_table.get_sym_table()
            global_map = self.symbol_table.get_sym_id_map()
            for col in ["cat", "name"]:
                self.traces[rank][col] = self.traces[rank][col].apply(
                    lambda idx: global_map[local_table[idx]]
                )

    def parse_multiple_ranks(
        self, ranks: List[int], use_multiprocessing: bool = True
    ) -> None:
        """
        Parse the trace for a given rank.

        Args:
            ranks (List[int]) : a list of integers representing the ranks of multiple trainers.
            use_multiprocessing (bool) : whether the parser using multiprocessing or not.
        """
        logger.debug(
            f"entering {sys._getframe().f_code.co_name}(ranks={ranks}, use_multiprocessing={use_multiprocessing})"
        )
        t0 = time.perf_counter()
        trace_paths = [self.trace_files[rank] for rank in ranks]
        logger.debug(f"trace_path={trace_paths}")
        local_symbol_tables: Dict[int, TraceSymbolTable] = {}
        if not use_multiprocessing:
            for rank in ranks:
                logger.debug(f"parsing trace for rank-{rank}")
                result = parse_trace_dataframe(self.trace_files[rank])
                self.meta_data[rank], self.traces[rank], local_symbol_tables[rank] = (
                    result[0],
                    result[1],
                    result[2],
                )
                self.symbol_table.add_symbols(local_symbol_tables[rank].get_sym_table())
            logger.debug(f"finished parsing for all {len(ranks)} ranks")
        else:
            num_procs = min(mp.cpu_count(), len(ranks))
            if len(ranks) <= 8:
                num_procs = min(len(ranks), mp.cpu_count())
            else:
                tracemalloc.start()
                parse_trace_dataframe(trace_paths[0])
                current, peak = tracemalloc.get_traced_memory()
                tracemalloc.stop()
                num_procs = get_mp_pool_size(peak, len(ranks))
            logger.debug(f"using {num_procs} processes for parsing.")

            with mp.get_context("fork").Pool(num_procs) as pool:
                results = pool.map(parse_trace_dataframe, trace_paths)
                pool.close()
                pool.join()
            logger.debug(f"finished parallel parsing using {num_procs} processes.")

            # collect the results
            for rank, result in zip(ranks, results):
                self.meta_data[rank], self.traces[rank], local_symbol_tables[rank] = (
                    result[0],
                    result[1],
                    result[2],
                )
                self.symbol_table.add_symbols(local_symbol_tables[rank].get_sym_table())

        # Now we update the IDs in the Dataframe using the global symbols table.
        global_map = self.symbol_table.get_sym_id_map()
        for rank in ranks:
            local_table = local_symbol_tables[rank].get_sym_table()
            for col in ["cat", "name"]:
                self.traces[rank][col] = self.traces[rank][col].apply(
                    lambda idx: global_map[local_table[idx]]
                )

        t1 = time.perf_counter()
        logger.debug(
            f"leaving {sys._getframe().f_code.co_name} duration={t1 - t0:.2f} seconds"
        )

    def parse_traces(
        self,
        max_ranks: int = -1,
        use_multiprocessing: bool = True,
    ) -> None:
        """
        Parse up to <max_rank> traces.

        Args:
            max_ranks (int): how many rank's traces to parse. Default value `-1` implies parsing all ranks.
            use_multiprocessing (bool) : whether the parser using multiprocessing or not.

        Effects:
            This function will parse the traces and save the parsed data into `self.traces`.
        """
        logger.debug(
            f"entering {sys._getframe().f_code.co_name}(max_ranks={max_ranks}, use_multiprocessing={use_multiprocessing})"
        )
        t0 = time.perf_counter()
        max_ranks = len(self.trace_files) if max_ranks == -1 else max_ranks
        ranks = sorted(self.trace_files.keys())[:max_ranks]
        logger.info(f"ranks={ranks}")

        if len(ranks) > 0:
            self.parse_multiple_ranks(ranks, use_multiprocessing and len(ranks) > 1)
            self.is_parsed = True
        else:
            logger.error("The list of ranks to be parsed is empty.")
            self.is_parsed = False
        t1 = time.perf_counter()
        logger.debug(
            f"leaving {sys._getframe().f_code.co_name} duration={t1 - t0:.2f} seconds"
        )

    def align_and_filter_trace(self):
        """
        Align the starting time across multiple ranks and filter events that belong to incomplete iterations.
        """
        self._align_all_ranks()
        self._filter_irrelevant_gpu_kernels()

    def get_trace(self, rank: int) -> pd.DataFrame:
        """
        Get the trace for a given rank.

        Args:
            rank (int) : the rank of the trainer whose trace is to be returned.

        Returns:
            The DataFrame for the given rank.

        Raises:
            ValueError when this Trace object doesn't have trace for the given rank.
        """
        if rank not in self.traces:
            logger.error(f"get_rank_trace - no trace for rank {rank}")
            raise ValueError
        return self.traces[rank]

    def get_all_traces(self) -> Dict[int, pd.DataFrame]:
        """
        Get the traces of all ranks.
        Returns:
            A dictionary with rank as key and its trace test_data as value.
        """
        return self.traces

    def get_raw_trace_for_one_rank(self, rank: int = 0) -> Dict[str, Any]:
        """
        Get raw trace content for one rank without filtering or compression to support writing back
        to a trace file.

        Args:
            rank (int) : the rank of the trainer whose trace is to be returned.

        Returns:
            The raw content of the trace file of the given rank.

        Raises:
            ValueError when this Trace object doesn't have trace for the given rank.
        """
        if rank not in self.trace_files:
            logger.error(f"get_rank_trace - no trace for rank {rank}")
            raise ValueError
        trace_filepath = self.trace_files[rank]
        return parse_trace_dict(trace_filepath)

    def write_raw_trace(self, output_file: str, trace_contents: Dict[str, Any]) -> None:
        with gzip.open(output_file, "wt") as fp:
            json.dump(trace_contents, fp, indent=2)

    def _normalize_trace_filenames(self) -> None:
        """
        Normalize the trace filenames so that a rank's trace file can be located by self.trace_files[rank] itself.
        """
        if not isinstance(self.trace_files, dict):
            logger.error(
                f"self.trace_files must be of type Dict[int, str]; got {type(self.trace_files)}"
            )
            raise ValueError(
                f"Expected trace_files to be Dict[int, str]; got {type(self.trace_files)}"
            )

        for rank in self.trace_files:
            filename = self.trace_files[rank]
            if not (filename.startswith(self.trace_path) or filename.startswith("/")):
                self.trace_files[rank] = os.path.join(self.trace_path, filename)

    def _validate_trace_files(self) -> bool:
        """
        Validate whether all trace files exist and are valid.

        Returns:
            success (bool) : True if all trace files in self.trace_files exist and are valid; False otherwise.
        """
        for rank, filepath in self.trace_files.items():
            if not (os.path.exists(filepath) and os.access(filepath, os.R_OK)):
                logger.error(
                    f"Trace file '{filepath}' doesn't exist or is not readable."
                )
                return False
            if not (filepath.endswith(".gz") or filepath.endswith(".json")):
                logger.error(
                    f"The postfix of trace file '{filepath}' is neither '.gz' or '.json'"
                )
                return False
        return True

    def _align_all_ranks(self) -> None:
        """
        Align dataframes for all ranks such that the earliest event starts at time 0.
        """
        self.min_ts = min(trace_df["ts"].min() for trace_df in self.traces.values())
        for rank, trace_df in self.traces.items():
            trace_df["ts"] = trace_df["ts"] - self.min_ts
            self.traces[rank] = trace_df

    def _filter_irrelevant_gpu_kernels(self) -> None:
        """
        Filter out GPU kernels that are not launched by the CPU kernels in the traced iterations.
        """
        sym_index = self.symbol_table.get_sym_id_map()
        profiler_steps = [v for k, v in sym_index.items() if "ProfilerStep" in k]

        def filter_gpu_kernels_for_one_rank(trace_df: pd.DataFrame) -> pd.DataFrame:
            cpu_kernels = trace_df[trace_df["stream"].eq(-1)]
            gpu_kernels = trace_df[trace_df["stream"].ne(-1)]
            last_profiler_start = cpu_kernels[cpu_kernels["name"].isin(profiler_steps)][
                "ts"
            ].max()
            cpu_kernels = cpu_kernels[cpu_kernels["ts"] < last_profiler_start]
            filtered_gpu_kernels = gpu_kernels.merge(
                cpu_kernels["correlation"], on="correlation", how="inner"
            )
            return pd.concat([filtered_gpu_kernels, cpu_kernels], axis=0)

        if not profiler_steps:
            logger.warning(
                "ProfilerStep not found in the trace. The analysis result may not be accurate."
            )
        elif len(profiler_steps) == 1:
            logger.warning(
                "There is only one iteration in the trace. The analysis result may not be accurate."
            )
        else:
            for rank, trace_df in self.traces.items():
                self.traces[rank] = filter_gpu_kernels_for_one_rank(trace_df)

    def convert_time_series_to_events(
        self, series: pd.DataFrame, counter_name: str, counter_col: str
    ) -> List[Dict[str, Any]]:
        """
        Takes time series data and convert it to Counter event format as per
        the Chrome Trace Format.
        See - https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/preview#heading=h.msg3086636uq

        @args
            series: is a dataframe with a minimum of pid, tid, ts.
                    (the id field is optional but use it to distinguish series with same name)
            counter_name (str): name of the counter.
            counter_col (str): name of the column used as the counter itself.

        Returns a list of json events that can be appended to the trace.
        """
        required_columns = ["pid", "ts", counter_col]
        if not set(required_columns).issubset(series.columns):
            logger.warning(
                "Time series dataframe does NOT contain required columns "
                f"{required_columns}, columns contained = {series.columns}"
            )
            return []

        events_df = series[required_columns].rename(columns={counter_col: "args"})
        events_df["ph"] = PHASE_COUNTER

        if "name" in series.columns:
            events_df["name"] = series["name"]
        else:
            events_df["name"] = counter_name
        if "id" in series.columns:
            events_df["id"] = series["id"]

        # add back ts delta
        events_df["ts"] = events_df["ts"] + self.min_ts

        def convert_to_args(value: int):
            return {counter_name: value}

        events_df.args = events_df.args.apply(convert_to_args)

        return events_df.to_dict("records")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from time import perf_counter
from typing import List, Optional

import pandas as pd
import plotly.express as px

from hta.common.trace import Trace, TraceSymbolTable
from hta.configs.config import logger


def plot_timeline(
    title: str, events: pd.DataFrame, ranks: Optional[List[int]] = None
) -> None:
    """
    Plot the timeline of events

    Args:
        title (str): a title for the timeline plot
        events (pd.DataFrame): a data contains the events to be plotted.
        ranks (List[int]) : a list of ranks whose events will be plotted.

    Return:
        None

    Raise:
        ValueError

    Notes:
        For the timeline to be properly drawn, the events DataFrame must contain the following columns:
        + calibrated_start_global
        + calibrated_end_global
        + task
        + label
    """
    t0 = perf_counter()
    must_have_columns: List[str] = [
        "calibrated_start_global",
        "calibrated_end_global",
        "task",
        "label",
    ]
    if not set(must_have_columns).issubset(set(events.columns)):
        raise ValueError(
            f"the events dataframe doesn't contain all required columns {must_have_columns}"
        )

    if ranks is None:
        if "rank" not in events.columns:
            ranks = []
        else:
            ranks = sorted(events["rank"].unique())

    addition_columns: List[str] = [
        "rank",
        "stream",
        "iteration",
        "s_cat",
        "cat",
        "s_name",
        "name",
        "dur",
    ]
    hover_data = sorted(list(set(events.columns).intersection(set(addition_columns))))

    sorted_tasks = sorted(events["task"].unique())
    fig = px.timeline(
        events,
        x_start="calibrated_start_global",
        x_end="calibrated_end_global",
        y="task",
        hover_data=hover_data,
        category_orders={"task": sorted_tasks},
        color="label",
        color_discrete_sequence=px.colors.qualitative.D3,
        width=1600,
        height=200 + 120 * len(ranks),
        title=title,
    )
    fig.show()

    t1 = perf_counter()
    logger.debug(f"Plotted timeline in {t1 - t0:.2f} seconds")


def _get_unique_values(
    df: pd.DataFrame, col: str, exclude_values: List[int] = [-1]
) -> List[int]:
    """Get the unique values for a given column <col> in the input Dataframe <df>

    Args:
        df (pd.DataFrame) : a DataFrame with trace events
        col (str) : a column name
        exclude_values (List[int]) : list of values to be excluded from the returned list

    Return:
        a list of unique values for df[col]
    """
    if col not in df:
        raise ValueError(f"{col} not in the DataFrame's columns: {df.columns}")
    unqiue_values = [e for e in df[col].unique().tolist() if e not in exclude_values]
    return sorted(unqiue_values)


def _simplify_name(name: str) -> str:
    """Simplify the name"""
    _patterns = [
        r"\(.+\)",
        r"<[^<]+>?",
        r"^void\s+",
        r"\s+",
        r"autograd::engine::evaluate_function",
    ]
    for pat in _patterns:
        name = re.sub(pat, "", name)
    return name if len(name) < 45 else name[:42] + "..."


def prepare_timeline_gpu_events(
    df: pd.DataFrame,
    symbol_table: TraceSymbolTable,
    ranks: Optional[List[int]] = None,
    iterations: Optional[List[int]] = None,
    streams: Optional[List[int]] = None,
    duration_threshold: int = 1000,
) -> pd.DataFrame:
    """
    Prepare GPU events for timeline analysis

    Args:
        df (pd.DataFrame) : an input DataFrame.
        symbol_table (TraceSymbolTable): input trace symbol table for mapping symbol id back to text.
        ranks (List[int]) : filter the input DataFrame with the given set of ranks; use all ranks if None.
        iterations (List[int]) : filter the input DataFrame with the given set of iteartions; use all iterations if None.
        streams (List[int]) : filter the input DataFrame with the given set of streams; use all streams if None.
        duration_threshold (int) : the minimum duration given for short kernels for them to be visible on the figure.
    Return:
        a DataFrame for selected GPU events for plot_timeline_px
    """
    t0 = perf_counter()
    required_columns: List[str] = ["iteration", "name", "cat", "rank", "stream", "ts"]
    if not set(required_columns).issubset(set(df.columns)):
        raise ValueError(
            f"columns {set(required_columns) - set(df.columns)} are not in input DataFrame"
        )

    if ranks is None:
        ranks = _get_unique_values(df, "rank")

    if iterations is None:
        iterations = _get_unique_values(df, "iteration")

    if streams is None:
        streams = _get_unique_values(df, "stream")

    events = df.loc[
        (df["rank"].isin(ranks))
        & (df["iteration"].isin(iterations))
        & (df["stream"].isin(streams))
    ].copy()

    sym_tab = symbol_table.get_sym_table()
    events["s_name"] = events["name"].apply(lambda i: sym_tab[i]).apply(_simplify_name)
    events["s_cat"] = events["cat"].apply(lambda i: sym_tab[i])
    events["calibrated_start_global"] = pd.to_datetime(events["ts"], unit="us")
    events["calibrated_end_global"] = pd.to_datetime(
        events["ts"]
        + events["dur"].apply(
            lambda x: x if x >= duration_threshold else duration_threshold
        ),
        unit="us",
    )

    events["label"] = events["s_name"]
    events["task"] = (
        "rank "
        + events["rank"].astype("str")
        + " stream "
        + events["stream"].astype("str")
    )
    events = events.sort_values(by=["rank", "stream", "ts"])

    t1 = perf_counter()
    logger.debug(
        f"Preprocessed events data for timeline visualization in {t1 - t0:.2f} seconds"
    )

    return events


def plot_timeline_gpu_kernels(
    title: str,
    df: pd.DataFrame,
    symbol_table: TraceSymbolTable,
    ranks: Optional[List[int]] = None,
    iterations: Optional[List[int]] = None,
    streams: Optional[List[int]] = None,
    duration_threshold: int = 1000,
) -> None:
    """
    Plot the timeline of selected GPU kernels in a DataFrame.

    Args:
        title (str): a title for the plot
        df: an input DataFrame.
        symbol_table: input trace symbol table for mapping symbol id back to text.
        ranks List[int]: filter the input DataFrame with the given set of ranks; use all ranks if None.
        iterations List[int]: filter the input DataFrame with the given set of iterations; use all iterations if None.
        streams List[int]: filter the input DataFrame with the given set of streams; use all streams if None.
        duration_threshold (int) : the minimum duration given for short kernels for them to be visible on the figure.
    """
    df_timeline = prepare_timeline_gpu_events(
        df, symbol_table, ranks, iterations, streams, duration_threshold
    )
    plot_timeline(title, df_timeline)


def plot_timeline_gpu_kernels_from_trace(
    title: str,
    trace_data: Trace,
    ranks: Optional[List[int]] = None,
    iterations: Optional[List[int]] = None,
    streams: Optional[List[int]] = None,
    duration_threshold: int = 1000,
) -> None:
    """
    Plot the timeline of selected GPU kernels in a Trace object.

    Args:
        title (str): a title for the timeline plot
        trace_data (Trace): a Trace object
        ranks (List[int]): filter the input DataFrame with the given set of ranks; use all ranks if None.
        iterations (List[int]): filter the input DataFrame with the given set of iterations; use all iterations if None.
        streams (List[int]): filter the input DataFrame with the given set of streams; use all streams if None.
        duration_threshold (int) : the minimum duration given for short kernels for them to be visible on the figure.
    """
    if ranks is not None:
        _ranks = list(set(trace_data.get_all_traces().keys()).intersection(set(ranks)))
    else:
        _ranks = list(trace_data.get_all_traces().keys())
    df = pd.concat(
        [trace_data.get_trace(r) for r in _ranks],
        axis=0,
        keys=_ranks,
        names=["rank", "idx"],
    ).reset_index()

    plot_timeline_gpu_kernels(
        title,
        df,
        trace_data.symbol_table,
        ranks,
        iterations,
        streams,
        duration_threshold,
    )
